"use strict";
/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `PostgresDriver` and related types declaration.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PostgresDriver = void 0;
const shared_1 = require("@cubejs-backend/shared");
const pg_1 = require("pg");
const moment = __importStar(require("moment"));
const base_driver_1 = require("@cubejs-backend/base-driver");
const QueryStream_1 = require("./QueryStream");
const GenericTypeToPostgres = {
    string: 'text',
    double: 'decimal',
    int: 'int8',
    // Revert mapping for internal pre-aggregations
    HLL_POSTGRES: 'hll',
};
const NativeTypeToPostgresType = {};
Object.entries(pg_1.types.builtins).forEach(([key, value]) => {
    NativeTypeToPostgresType[value] = key;
});
// pg-types lacks the default `unknown` type since it's a pseudo-type
NativeTypeToPostgresType['705'] = 'UNKNOWN';
const PostgresToGenericType = {
    // bpchar (“blank-padded char”, the internal name of the character data type)
    bpchar: 'varchar',
    // Numeric is an alias
    numeric: 'decimal',
    // External mapping
    hll: 'HLL_POSTGRES',
};
const timestampDataTypes = [
    // @link TypeId.DATE
    1082,
    // @link TypeId.TIMESTAMP
    1114,
    // @link TypeId.TIMESTAMPTZ
    1184
];
const timestampTypeParser = (val) => moment.utc(val).format(moment.HTML5_FMT.DATETIME_LOCAL_MS);
const hllTypeParser = (val) => Buffer.from(
// Postgres uses prefix as \x for encoding
val.slice(2), 'hex').toString('base64');
/**
 * Postgres driver class.
 */
class PostgresDriver extends base_driver_1.BaseDriver {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 2;
    }
    enabled = false;
    pool;
    config;
    /**
     * Class constructor.
     */
    constructor(config = {}) {
        super({
            testConnectionTimeout: config.testConnectionTimeout,
        });
        const dataSource = config.dataSource ||
            (0, shared_1.assertDataSource)('default');
        this.pool = new pg_1.Pool({
            idleTimeoutMillis: 30000,
            max: config.maxPoolSize ||
                (0, shared_1.getEnv)('dbMaxPoolSize', { dataSource }) ||
                8,
            host: (0, shared_1.getEnv)('dbHost', { dataSource }),
            database: (0, shared_1.getEnv)('dbName', { dataSource }),
            port: (0, shared_1.getEnv)('dbPort', { dataSource }),
            user: (0, shared_1.getEnv)('dbUser', { dataSource }),
            password: (0, shared_1.getEnv)('dbPass', { dataSource }),
            ssl: this.getSslOptions(dataSource),
            ...config
        });
        this.pool.on('error', (err) => {
            console.log(`Unexpected error on idle client: ${err.stack || err}`); // TODO
        });
        this.config = {
            ...this.getInitialConfiguration(dataSource),
            executionTimeout: (0, shared_1.getEnv)('dbQueryTimeout', { dataSource }),
            exportBucketCsvEscapeSymbol: (0, shared_1.getEnv)('dbExportBucketCsvEscapeSymbol', { dataSource }),
            ...config,
        };
        this.enabled = true;
    }
    primaryKeysQuery(conditionString) {
        return `SELECT
      columns.table_schema as ${this.quoteIdentifier('table_schema')},
      columns.table_name as ${this.quoteIdentifier('table_name')},
      columns.column_name as ${this.quoteIdentifier('column_name')}
    FROM information_schema.table_constraints tc
    JOIN information_schema.constraint_column_usage AS ccu USING (constraint_schema, constraint_name)
    JOIN information_schema.columns AS columns ON columns.table_schema = tc.constraint_schema
      AND tc.table_name = columns.table_name AND ccu.column_name = columns.column_name
    WHERE constraint_type = 'PRIMARY KEY' AND columns.table_schema NOT IN ('pg_catalog', 'information_schema', 'mysql', 'performance_schema', 'sys', 'INFORMATION_SCHEMA')${conditionString ? ` AND (${conditionString})` : ''}`;
    }
    foreignKeysQuery(conditionString) {
        return `SELECT
        tc.table_schema as ${this.quoteIdentifier('table_schema')},
        tc.table_name as ${this.quoteIdentifier('table_name')},
        kcu.column_name as ${this.quoteIdentifier('column_name')},
        columns.table_name as ${this.quoteIdentifier('target_table')},
        columns.column_name as ${this.quoteIdentifier('target_column')}
      FROM
        information_schema.table_constraints AS tc
      JOIN information_schema.key_column_usage AS kcu
        ON tc.constraint_name = kcu.constraint_name
      JOIN information_schema.constraint_column_usage AS columns
        ON columns.constraint_name = tc.constraint_name
      WHERE
         constraint_type = 'FOREIGN KEY'
         AND ${this.getColumnNameForSchemaName()} NOT IN ('pg_catalog', 'information_schema', 'mysql', 'performance_schema', 'sys', 'INFORMATION_SCHEMA')
         ${conditionString ? ` AND (${conditionString})` : ''}
    `;
    }
    /**
     * The easiest way how to add additional configuration from env variables, because
     * you cannot call method in RedshiftDriver.constructor before super.
     */
    getInitialConfiguration(
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    dataSource) {
        return {
            readOnly: true,
        };
    }
    getTypeParser = (dataTypeID, format) => {
        const isTimestamp = timestampDataTypes.includes(dataTypeID);
        if (isTimestamp) {
            return timestampTypeParser;
        }
        const typeName = this.getPostgresTypeForField(dataTypeID);
        if (typeName === 'hll') {
            // We are using base64 encoding as main format for all HLL sketches, but in pg driver it uses binary encoding
            return hllTypeParser;
        }
        const parser = pg_1.types.getTypeParser(dataTypeID, format);
        return (val) => parser(val);
    };
    /**
     * It's not possible to detect user defined types via constant oids
     * For example HLL extensions is using CREATE TYPE HLL which will generate a new pg_type with different oids
     */
    userDefinedTypes = null;
    getPostgresTypeForField(dataTypeID) {
        if (dataTypeID in NativeTypeToPostgresType) {
            return NativeTypeToPostgresType[dataTypeID].toLowerCase();
        }
        if (this.userDefinedTypes && dataTypeID in this.userDefinedTypes) {
            return this.userDefinedTypes[dataTypeID].toLowerCase();
        }
        return null;
    }
    async testConnection() {
        try {
            await this.pool.query('SELECT $1::int AS number', ['1']);
        }
        catch (e) {
            if (e.toString().indexOf('no pg_hba.conf entry for host') !== -1) {
                throw new Error(`Please use CUBEJS_DB_SSL=true to connect: ${e.toString()}`);
            }
            throw e;
        }
    }
    async loadUserDefinedTypes(conn) {
        if (!this.userDefinedTypes) {
            // Postgres enum types defined as typcategory = 'E' these can be assumed
            // to be of type varchar for the drivers purposes.
            // TODO: if full implmentation the constraints can be looked up via pg_enum
            // https://www.postgresql.org/docs/9.1/catalog-pg-enum.html
            const customTypes = await conn.query(`SELECT
            oid,
            CASE
                WHEN typcategory = 'E' THEN 'varchar'
                ELSE typname
            END
        FROM
            pg_type
        WHERE
            typcategory in ('U', 'E')`, []);
            this.userDefinedTypes = customTypes.rows.reduce((prev, current) => ({ [current.oid]: current.typname, ...prev }), {});
        }
    }
    async prepareConnection(conn, options = {
        executionTimeout: this.config.executionTimeout ? (this.config.executionTimeout) * 1000 : 600000
    }) {
        await conn.query(`SET TIME ZONE '${this.config.storeTimezone || 'UTC'}'`);
        await conn.query(`SET statement_timeout TO ${options.executionTimeout}`);
        await this.loadUserDefinedTypes(conn);
    }
    mapFields(fields) {
        return fields.map((f) => {
            const postgresType = this.getPostgresTypeForField(f.dataTypeID);
            if (!postgresType) {
                throw new Error(`Unable to detect type for field "${f.name}" with dataTypeID: ${f.dataTypeID}`);
            }
            return ({
                name: f.name,
                type: this.toGenericType(postgresType)
            });
        });
    }
    async stream(query, values, { highWaterMark }) {
        PostgresDriver.checkValuesLimit(values);
        const conn = await this.pool.connect();
        try {
            await this.prepareConnection(conn);
            const queryStream = new QueryStream_1.QueryStream(query, values, {
                types: {
                    getTypeParser: this.getTypeParser,
                },
                highWaterMark
            });
            const rowStream = await conn.query(queryStream);
            const fields = await rowStream.fields();
            return {
                rowStream,
                types: this.mapFields(fields),
                release: async () => {
                    await conn.release();
                }
            };
        }
        catch (e) {
            await conn.release();
            throw e;
        }
    }
    static checkValuesLimit(values) {
        // PostgreSQL protocol allows sending up to 65535 params in a single bind message
        // See https://github.com/postgres/postgres/blob/REL_16_0/src/backend/tcop/postgres.c#L1698-L1708
        // See https://github.com/postgres/postgres/blob/REL_16_0/src/backend/libpq/pqformat.c#L428-L431
        // But 'pg' module does not check for params count, and ends up sending incorrect bind message
        // See https://github.com/brianc/node-postgres/blob/92cb640fd316972e323ced6256b2acd89b1b58e0/packages/pg-protocol/src/serializer.ts#L155
        // See https://github.com/brianc/node-postgres/blob/92cb640fd316972e323ced6256b2acd89b1b58e0/packages/pg-protocol/src/buffer-writer.ts#L32-L37
        const length = (values?.length ?? 0);
        if (length >= 65536) {
            throw new Error(`PostgreSQL protocol does not support more than 65535 parameters, but ${length} passed`);
        }
    }
    async queryResponse(query, values) {
        PostgresDriver.checkValuesLimit(values);
        const conn = await this.pool.connect();
        try {
            await this.prepareConnection(conn);
            const res = await conn.query({
                text: query,
                values: values || [],
                types: {
                    getTypeParser: this.getTypeParser,
                },
            });
            return res;
        }
        finally {
            await conn.release();
        }
    }
    async createTable(quotedTableName, columns) {
        if (quotedTableName.length > 63) {
            throw new Error('PostgreSQL can not work with table names longer than 63 symbols. ' +
                `Consider using the 'sqlAlias' attribute in your cube definition for ${quotedTableName}.`);
        }
        return super.createTable(quotedTableName, columns);
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    async query(query, values, options) {
        const result = await this.queryResponse(query, values);
        return result.rows;
    }
    async downloadQueryResults(query, values, options) {
        if (options.streamImport) {
            return this.stream(query, values, options);
        }
        const res = await this.queryResponse(query, values);
        return {
            rows: res.rows,
            types: this.mapFields(res.fields),
        };
    }
    toGenericType(columnType) {
        if (columnType in PostgresToGenericType) {
            return PostgresToGenericType[columnType];
        }
        return super.toGenericType(columnType);
    }
    readOnly() {
        return !!this.config.readOnly;
    }
    async uploadTableWithIndexes(table, columns, tableData, indexesSql) {
        if (!tableData.rows) {
            throw new Error(`${this.constructor} driver supports only rows upload`);
        }
        await this.createTable(table, columns);
        try {
            await this.query(`INSERT INTO ${table}
      (${columns.map(c => this.quoteIdentifier(c.name)).join(', ')})
      SELECT * FROM UNNEST (${columns.map((c, columnIndex) => `${this.param(columnIndex)}::${this.fromGenericType(c.type)}[]`).join(', ')})`, columns.map(c => tableData.rows.map(r => r[c.name])));
            for (let i = 0; i < indexesSql.length; i++) {
                const [query, p] = indexesSql[i].sql;
                await this.query(query, p);
            }
        }
        catch (e) {
            await this.dropTable(table);
            throw e;
        }
    }
    async release() {
        if (this.enabled) {
            this.pool.end();
            this.enabled = false;
        }
    }
    param(paramIndex) {
        return `$${paramIndex + 1}`;
    }
    fromGenericType(columnType) {
        return GenericTypeToPostgres[columnType] || super.fromGenericType(columnType);
    }
    capabilities() {
        return {
            incrementalSchemaLoading: true,
        };
    }
}
exports.PostgresDriver = PostgresDriver;
//# sourceMappingURL=PostgresDriver.js.map